/* $USAGI: pfkey_v2_msg_flow.c,v 1.16 2001/09/28 16:52:40 mk Exp $ */

/*
 * Copyright (C) 2001 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * This is a parse routine for messages SADB_EXT_ADDFLOW and SADB_EXT_DELFLOW.
 * We use FreeS/WAN's user land routine for manipulating SA and Policy.
 * These are FreeS/WAN specific.
 */

#ifdef PFKEY_DEBUG
#define DEBUG 1
#endif

#ifdef MODULE
#include <linux/module.h>
#  ifdef MODVERSIONS
#  include <linux/modversions.h>
#  endif
#endif

#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/random.h>

#include <net/pfkeyv2.h>
#include <net/pfkey.h>
#include <net/sadb.h>
#include <net/spd.h>

#include "pfkey_v2_msg.h"
#include "sadb_utils.h" /* esptoa authtoa */
			/* sockaddrtoa ultoa */
#include "spd_utils.h"
#define BUFSIZE 64

int sadb_msg_addflow_parse(struct sock *sk, struct sadb_msg* msg, struct sadb_msg **reply)
{
	int error = 0;
	char buf[BUFSIZE];
	struct sadb_ext *ext_msgs[SADB_EXT_MAX+1];
	struct sadb_sa *sa;
	struct sadb_address *src = NULL;
	struct sadb_address *dst = NULL;
	struct sadb_address *sflow = NULL;
	struct sadb_address *dflow = NULL;
	struct ipsec_sp *policy = NULL;
	struct sa_index sa_idx;
	struct selector selector;


	if (!msg) {
		pr_debug("sadb_msg_addflow_parse: msg is null\n");
		error = -EINVAL;
		goto err;
	}

	memset(ext_msgs, 0, sizeof(ext_msgs));
	error = sadb_msg_detect_ext(msg, ext_msgs);

	if (ext_msgs[SADB_EXT_SA] &&
		ext_msgs[SADB_EXT_ADDRESS_SRC] &&
		ext_msgs[SADB_EXT_ADDRESS_DST] &&
		ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW] &&
		ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW])
	{
		sa_index_init(&sa_idx);

		sa  = (struct sadb_sa*)ext_msgs[SADB_EXT_SA];
		pr_debug("sadb_msg_addflow_parse: sa.spi=0x%x\n", ntohl(sa->sadb_sa_spi));
		pr_debug("sadb_msg_addflow_parse: sa.auth=%u\n", ntohs(sa->sadb_sa_auth));
		pr_debug("sadb_msg_addflow_parse: sa.encrypt=%u\n", ntohs(sa->sadb_sa_encrypt));

		sa_idx.ipsec_proto = msg->sadb_msg_satype;
		sa_idx.spi = sa->sadb_sa_spi;

		src = (struct sadb_address*)ext_msgs[SADB_EXT_ADDRESS_SRC];
		dst = (struct sadb_address*)ext_msgs[SADB_EXT_ADDRESS_DST];
		sflow = (struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW];
		dflow = (struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW];

		if (sflow->sadb_address_proto == dflow->sadb_address_proto) {
			selector.proto = sflow->sadb_address_proto;
		} else {
			error = -EINVAL;
			goto err;
		}
		memset(buf, 0, BUFSIZE);
		sadb_address_to_sockaddr( dst, (struct sockaddr*)&(sa_idx.dst));
		sockaddrtoa((struct sockaddr*)&(sa_idx.dst), buf, BUFSIZE);
		pr_debug("sadb_msg_addflow_parse: dst=%s\n", buf);

		memset(buf, 0, BUFSIZE);
		sadb_address_to_sockaddr( sflow, (struct sockaddr*)&(selector.src));
		sockaddrtoa((struct sockaddr*)&(selector.src), buf, BUFSIZE);
		pr_debug("sadb_msg_addflow_parse: sflow=%s\n", buf);

		memset(buf, 0, BUFSIZE);
		sadb_address_to_sockaddr( dflow, (struct sockaddr*)&(selector.dst));
		sockaddrtoa((struct sockaddr*)&(selector.dst), buf, BUFSIZE);
		pr_debug("sadb_msg_addflow_parse: dflow=%s\n", buf);

		selector.prefixlen_s = sflow->sadb_address_prefixlen;
		pr_debug("sadb_msg_addflow_parse: prefixlen_s=%d\n", selector.prefixlen_s);
		selector.prefixlen_d = dflow->sadb_address_prefixlen;
		pr_debug("sadb_msg_addflow_parse: prefixlen_d=%d\n", selector.prefixlen_d);

		error = spd_find_by_selector(&selector, &policy, SPD_WITH_WLOCK);

		pr_debug("sadb_msg_addflow_parse: spd_find error = %d\n", error);

		if (error == -ESRCH) {

			/* It's new one. I append it into spd_list */

			policy = ipsec_sp_kmalloc();
			if (!policy) {
				pr_debug("sadb_msg_addflow_parse: ipsec_sp_kmalloc faild\n");
				error = -ENOMEM;
				goto err;
			}
			memcpy(&(policy->selector), &selector, sizeof(struct selector));

			error = sa_list_append(&(policy->sa_index_list), &sa_idx);
			if (error)
				goto err;
			dump_ipsec_sp(policy);
			error = spd_append(policy);
			if (error)
				goto err;
			error = ipsec_sp_kfree(&policy);

		} else if (error == -EEXIST) {

			/* It has already been in spd_list, I append sa_index into it's sa_list */
			pr_debug("pfkey_msg_addflow_parse: policy=%p\n", policy);
			error = sa_list_append(&(policy->sa_index_list), &sa_idx);
			if (error) {
				SPD_WUNLOCK(policy);
				goto err;
			}
			dump_ipsec_sp(policy);
			SPD_WUNLOCK(policy);

		} else {
			/* Something is wrong */
			pr_debug("sadb_msg_addflow_parse: spd_find_by_selector faild\n");
			goto err;
		}

	}

//	sadb_msg_reply_build(msg, ext_msgs, reply);
	return error;
err:
//	sadb_msg_reply_build2(msg, error, reply);
	return error;

}

int sadb_msg_delflow_parse(struct sock *sk, struct sadb_msg* msg, struct sadb_msg **reply)
{
	int error = 0;
	struct sadb_ext *ext_msgs[SADB_EXT_MAX+1];
	struct sadb_address *sflow = NULL;
	struct sadb_address *dflow = NULL;
	struct sa_index sa_idx;
	struct selector selector;


	if (!msg) {
		pr_debug("sadb_msg_delflow_parse: msg is null\n");
		error = -EINVAL;
		goto err;
	}

	memset(ext_msgs, 0, sizeof(ext_msgs));
	error = sadb_msg_detect_ext(msg, ext_msgs);

	if (ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW] &&
		ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW])
	{
		sa_index_init(&sa_idx);

		sflow = (struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW];
		dflow = (struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW];

		if (sflow->sadb_address_proto == dflow->sadb_address_proto) {
			selector.proto = sflow->sadb_address_proto;
		} else {
			error = -EINVAL;
			goto err;
		}	

		sadb_address_to_sockaddr( sflow, (struct sockaddr*)&(selector.src));
		sadb_address_to_sockaddr( dflow, (struct sockaddr*)&(selector.dst));
		selector.prefixlen_s = sflow->sadb_address_prefixlen;
		selector.prefixlen_d = dflow->sadb_address_prefixlen;

		error = spd_remove(&selector);

		if (error) {
			pr_debug("sadb_msg_delflow_parse: spd_remove failed\n");
			goto err;
		}

	}

//	sadb_msg_reply_build(msg, ext_msgs, reply);
	return error;
err:
//	sadb_msg_reply_build2(msg, error, reply);
	return error;

}



